import { columns } from "../modules/chess-board/models";
import { FENConverter } from "./FENConverter";
import { Color, Coords, FENChar, CheckState, LastMove, SafeSquares, MoveType, MoveList, GameHistory } from "./models";
import { Bishop } from "./pieces/bishop";
import { King } from "./pieces/king";
import { Knight } from "./pieces/knight";
import { Pawn } from "./pieces/pawn";
import { Piece } from "./pieces/piece";
import { Queen } from "./pieces/queen";
import { Rook } from "./pieces/rook";

// Implement ChessBoard class
export class ChessBoard {

    private static chessBoardSize = 8;

    private chessBoard: (Piece | null)[][];
    private _playerColor = Color.White;
    private _safeSquares: SafeSquares;
    private _lastMove: LastMove | undefined
    private _checkState: CheckState = { isInChecked: false };
    private fiftyMoveRuleCounter: number = 0;


    private _isGameOver: boolean = false;
    private _gameOverMessage: string | undefined;

    private fullNumberOfMoves: number = 1;

    private _boardAsFEN: string = FENConverter.initalPosition;
    private FENConverter = new FENConverter();

    private _moveList: MoveList = [];
    private _gameHistory: GameHistory;

    // Threefold repetition
    private threeFoldRepetitionDictionary = new Map<string, number>();
    private threeFoldRepetitionFlag: boolean = false;

    constructor() {
        this.chessBoard = [
            [
                new Rook(Color.White),
                new Knight(Color.White),
                new Bishop(Color.White),
                new Queen(Color.White),
                new King(Color.White),
                new Bishop(Color.White),
                new Knight(Color.White),
                new Rook(Color.White)
            ],
            [
                new Pawn(Color.White),
                new Pawn(Color.White),
                new Pawn(Color.White),
                new Pawn(Color.White),
                new Pawn(Color.White),
                new Pawn(Color.White),
                new Pawn(Color.White),
                new Pawn(Color.White)
            ],
            new Array(8).fill(null),
            new Array(8).fill(null),
            new Array(8).fill(null),
            new Array(8).fill(null),
            [
                new Pawn(Color.Black),
                new Pawn(Color.Black),
                new Pawn(Color.Black),
                new Pawn(Color.Black),
                new Pawn(Color.Black),
                new Pawn(Color.Black),
                new Pawn(Color.Black),
                new Pawn(Color.Black)
            ],
            [
                new Rook(Color.Black),
                new Knight(Color.Black),
                new Bishop(Color.Black),
                new Queen(Color.Black),
                new King(Color.Black),
                new Bishop(Color.Black),
                new Knight(Color.Black),
                new Rook(Color.Black)
            ],
        ];
        this._safeSquares = this.findSafeSquares();
 
        this._gameHistory = [{ 
            board: this.chessBoardView, 
            lastMove: this._lastMove, 
            checkState: this._checkState }];
    }

    public get playerColor(): Color {
        return this._playerColor;
    }

    public get chessBoardView(): (FENChar | null)[][] {
        return this.chessBoard.map(row =>
            row.map(piece => piece instanceof Piece ? piece.FENChar : null)
        );
    }

    public static isSquareDark(x: number, y: number): boolean {
        return x % 2 === 0 && y % 2 === 0 || x % 2 === 1 && y % 2 === 1;
    }

    public get safeSquares(): SafeSquares {
        return this._safeSquares;
    }

    public get lastMove(): LastMove | undefined {
        return this._lastMove;
    }

    public get checkState(): CheckState {
        return this._checkState;
    }

    public get isGameOver(): boolean {
        return this._isGameOver;
    }

    public get gameOverMessage(): string | undefined {
        return this._gameOverMessage;
    }

    public get boardAsFEN(): string {
        return this._boardAsFEN;
    }

    public get moveList(): MoveList {
        return this._moveList;
    }

    public get gameHistory(): GameHistory {
        return this._gameHistory;
    }

    private areCoordsValid(coords: { x: number, y: number }): boolean {
        return coords.x >= 0 && coords.y >= 0 && coords.x < ChessBoard.chessBoardSize && coords.y < ChessBoard.chessBoardSize;
    }

    public isInCheck(playerColor: Color, checkingCurrentPosition: boolean): boolean {
        // Iterate over the board
        for (let row = 0; row < ChessBoard.chessBoardSize; row++) {
            for (let col = 0; col < ChessBoard.chessBoardSize; col++) {
                // Bail if square not being occupied or the same color
                const piece = this.chessBoard[row][col];
                if (!piece || piece.color === playerColor) continue;

                // Check the piece's moves
                for (const direction of piece.directions) {
                    const coords = { x: row + direction.x, y: col + direction.y };

                    // Bail if coords are invalid
                    if (!this.areCoordsValid(coords)) continue;

                    // Check if piece instance of Pawn, King, or Knight
                    if (piece instanceof Pawn || piece instanceof King || piece instanceof Knight) {

                        // Bail if pawn and not diagonal
                        if (piece instanceof Pawn && direction.y === 0) continue;

                        const targetPiece = this.chessBoard[coords.x][coords.y];
                        // If the target piece is a King and not player color return true
                        if (targetPiece instanceof King && targetPiece.color === playerColor) {
                            if (checkingCurrentPosition) this._checkState = { isInChecked: true, coords: { x: coords.x, y: coords.y } };
                            return true;
                        }
                    }
                    else {
                        // Check if piece is a Bishop, Rook, or Queen
                        while (this.areCoordsValid(coords)) {

                            // If the target piece is a King and not player color return true
                            const targetPiece = this.chessBoard[coords.x][coords.y];
                            if (targetPiece instanceof King && targetPiece.color === playerColor) {
                                if (checkingCurrentPosition) this._checkState = { isInChecked: true, coords: { x: coords.x, y: coords.y } };
                                return true;
                            }

                            // If the target piece is not null, break
                            if (targetPiece !== null) break;


                            // Increment coords     
                            coords.x += direction.x;
                            coords.y += direction.y;
                        }
                    }

                }
            }

        }

        if (checkingCurrentPosition) this._checkState = { isInChecked: false };

        return false;
    }

    private isPositionSafeAfterMove(from: { x: number, y: number }, to: { x: number, y: number }): boolean {

        const piece = this.chessBoard[from.x][from.y];
        if (!piece) return false;

        const newPiece: Piece | null = this.chessBoard[to.x][to.y];
        // we cant put piece on a square that already contains piece of the same square
        if (newPiece && newPiece.color === piece.color) return false;

        // simulate position
        this.chessBoard[from.x][from.y] = null;
        this.chessBoard[to.x][to.y] = piece;

        const isPositionSafe: boolean = !this.isInCheck(piece.color, false);

        // restore position back
        this.chessBoard[from.x][from.y] = piece;
        this.chessBoard[to.x][to.y] = newPiece;

        return isPositionSafe;
    }

    private findSafeSquares(): SafeSquares {
        const safeSquares: SafeSquares = new Map<string, Coords[]>();

        const debugPiece = undefined;
        const debugX = 10;
        const debugY = 4;
        for (let x = 0; x < ChessBoard.chessBoardSize; x++) {
            for (let y = 0; y < ChessBoard.chessBoardSize; y++) {

                const piece: Piece | null = this.chessBoard[x][y];
                if (x === debugX && y === debugY) console.log('=========> findSafeSquares:Piece Started for coords', x, y, piece)
                if (!piece || piece.color !== this._playerColor) {
                    continue;
                }

                // console.log('findSafeSquares:Piece Started for coords', x, y, piece)

                const pieceSafeSquares: Coords[] = [];

                if (x === debugX && y === debugY) console.log('=========> findSafeSquares:Iterating over directions', x, y, piece)
                for (const { x: dx, y: dy } of piece.directions) {
                    let newX: number = x + dx;
                    let newY: number = y + dy;

                    if (!this.areCoordsValid({ x: newX, y: newY })) {
                        if (piece.FENChar === debugPiece) console.log('findSafeSquares:Coords not valid 1', newX, newY)
                        if (piece.FENChar === debugPiece) console.log('findSafeSquares:Bailing 1', piece)
                        continue;
                    }

                    let newPiece: Piece | null = this.chessBoard[newX][newY];
                    if (newPiece && newPiece.color === piece.color) {
                        if (piece.FENChar === debugPiece) console.log('findSafeSquares:Bailing 2', piece)
                        continue;
                    }

                    if (piece.FENChar === debugPiece) console.log('Found a ', piece.FENChar)

                    // need to restrict pawn moves in certain directions
                    if (piece instanceof Pawn) {
                        // cant move pawn two squares straight if there is piece infront of him
                        if (dx === 2 || dx === -2) {
                            if (newPiece) {
                                if (piece.FENChar === debugPiece) console.log('findSafeSquares:Bailing 3', piece)
                                continue;
                            }
                            if (this.chessBoard[newX + (dx === 2 ? -1 : 1)][newY]) {
                                if (piece.FENChar === debugPiece) console.log('findSafeSquares:Bailing 4', piece)
                                continue;
                            }
                        }

                        // cant move pawn one square straight if piece is infront of him
                        if ((dx === 1 || dx === -1) && dy === 0 && newPiece) {
                            if (piece.FENChar === debugPiece) console.log('findSafeSquares:Bailing 5', piece)
                            continue;
                        }

                        // cant move pawn diagonally if there is no piece, or piece has same color as pawn
                        if ((dy === 1 || dy === -1) && (!newPiece || piece.color === newPiece.color)) {
                            if (piece.FENChar === debugPiece) console.log('findSafeSquares:Bailing 6', piece)
                            continue;
                        }
                    }

                    if (piece instanceof Pawn || piece instanceof Knight || piece instanceof King) {
                        if (this.isPositionSafeAfterMove({ x, y }, { x: newX, y: newY })) {
                            pieceSafeSquares.push({ x: newX, y: newY });
                            if (piece.FENChar === debugPiece) console.log('findSafeSquares:NOT Bailing 7', pieceSafeSquares.length)
                            if (piece.FENChar === debugPiece) console.log('findSafeSquares:NOT Bailing 7', piece)
                        }
                        else {
                            if (piece.FENChar === debugPiece) console.log('findSafeSquares:Bailing 7', piece)
                        }
                    }
                    else {
                        while (this.areCoordsValid({ x: newX, y: newY })) {
                            newPiece = this.chessBoard[newX][newY];
                            if (newPiece && newPiece.color === piece.color) break;

                            if (this.isPositionSafeAfterMove({ x, y }, { x: newX, y: newY }))
                                pieceSafeSquares.push({ x: newX, y: newY });

                            if (newPiece !== null) {
                                if (piece.FENChar === debugPiece) console.log('findSafeSquares:Bailing 8', piece)
                                break;
                            }

                            newX += dx;
                            newY += dy;
                        }
                    }
                }
                if (x === debugX && y === debugY) console.log('=========> findSafeSquares:Completed Iterating over directions', x, y, piece)

                // Check if the piece is a King and can castle
                if (x === debugX && y === debugY) console.log('=========> findSafeSquares:About to check for king', x, y, piece)

                // Handle Castling    
                if (piece instanceof King) {
                    console.log('findSafeSquares:Checking for castling x is ', x, y)
                    if (this.canCastle(piece, true)) {
                        console.log('findSafeSquares:Pushing a', x, 6)
                        pieceSafeSquares.push({ x, y: 6 })
                    };
                    if (this.canCastle(piece, false)) {
                        console.log('findSafeSquares:Pushing b', x, 2)
                        pieceSafeSquares.push({ x, y: 2 });
                    }
                }
                // Handling En Passant
                if (piece instanceof Pawn) {
                    if (this.canCaptureEnPassant(piece, { x, y })) {
                        console.log('findSafeSquares:Pushing a', x - 1, y + 1)
                        pieceSafeSquares.push({ x: x + (piece.color === Color.White ? 1 : -1), y: this._lastMove!.from.y });
                    }
                }


                if (pieceSafeSquares.length) {
                    if (piece.FENChar === debugPiece) console.log('Setting at ', x, y, pieceSafeSquares)
                    safeSquares.set(x + "," + y, pieceSafeSquares);

                }
                if (x === debugX && y === debugY) console.log('=========> findSafeSquares:Piece Finished for coords', x, y, piece)
            }
        }

        return safeSquares;
    }

    public canCaptureEnPassant(pawn: Pawn, pawnCoords: Coords): boolean {

        if (!this._lastMove) return false;
        const { piece, from: lastFrom, to: LastTo } = this._lastMove;

        if (
            !(piece instanceof Pawn) ||
            pawn.color !== this._playerColor ||
            Math.abs(LastTo.x - lastFrom.x) !== 2 ||
            pawnCoords.x !== LastTo.x ||
            Math.abs(pawnCoords.y - LastTo.y) !== 1
        ) return false;

        const pawnNewPositionX: number = pawnCoords.x + (pawn.color === Color.White ? 1 : -1);
        const pawnNewPositionY: number = LastTo.y;

        this.chessBoard[LastTo.x][LastTo.y] = null;
        this.chessBoard[LastTo.x][LastTo.y] = piece;

        const isPositionSafe: boolean = this.isPositionSafeAfterMove({ x: pawnCoords.x, y: pawnCoords.y }, { x: pawnNewPositionX, y: pawnNewPositionY });

        return isPositionSafe;
    }

    public canCastle(king: King, kingSideCastle: boolean): boolean {
        const debugOn = false;
        const debugCollor = Color.White;
        const debugKingSideCastle = true;

        if (debugOn && king.color === debugCollor && kingSideCastle === debugKingSideCastle) console.log('==========>In canCastle', king, kingSideCastle)
        if (king.hasMoved) {
            if (debugOn && king.color === debugCollor && kingSideCastle === debugKingSideCastle) console.log('canCastle: Bailing king has moved')
            return false;
        }

        // Set the king position based on color
        const kingPositionX: number = king.color === Color.White ? 0 : 7;
        const kingPositionY: number = 4;

        // Set the rook position based on color
        const rookPositionX: number = kingPositionX;
        const rookPositionY: number = kingSideCastle ? 7 : 0;

        // Bail if Piece is not a rook, or it has moved, or the player is in check
        const rook: Piece | null = this.chessBoard[rookPositionX][rookPositionY];
        if (!(rook instanceof Rook) || rook.hasMoved || this._checkState.isInChecked) {
            if (debugOn && king.color === debugCollor && kingSideCastle === debugKingSideCastle) console.log('canCastle: Bailing. Not Rook or has Moved, or is in check Bailing 1', rook)
            return false;
        }

        // Calculate the next king position
        const firstNextKingPositionY: number = kingPositionY + (kingSideCastle ? 1 : -1);
        const secondNextKingPositionY: number = kingPositionY + (kingSideCastle ? 2 : -2);

        if (this.chessBoard[kingPositionX][firstNextKingPositionY] || this.chessBoard[kingPositionX][secondNextKingPositionY]) {
            if (debugOn && king.color === debugCollor && kingSideCastle === debugKingSideCastle) console.log('canCastle: Bailing 2a', this.chessBoard[kingPositionX][firstNextKingPositionY])
            if (debugOn && king.color === debugCollor && kingSideCastle === debugKingSideCastle) console.log('canCastle: Bailing 2b', this.chessBoard[kingPositionX][secondNextKingPositionY])
            if (debugOn && king.color === debugCollor && kingSideCastle === debugKingSideCastle) console.log('canCastle: Bailing 2c', kingPositionX, firstNextKingPositionY)
            if (debugOn && king.color === debugCollor && kingSideCastle === debugKingSideCastle) console.log('canCastle: Bailing 2d', kingPositionX, secondNextKingPositionY)
            if (debugOn && king.color === debugCollor && kingSideCastle === debugKingSideCastle) console.log('canCastle: Bailing 2')
            return false;
        }

        if (!kingSideCastle && this.chessBoard[kingPositionX][1]) {
            if (debugOn && king.color === debugCollor) console.log('canCastle: Bailing 3')
            return false;
        }

        const isPositionSafeA = this.isPositionSafeAfterMove({ x: kingPositionX, y: kingPositionY }, { x: kingPositionX, y: firstNextKingPositionY })
        const isPositionSafeB = this.isPositionSafeAfterMove({ x: kingPositionX, y: kingPositionY }, { x: kingPositionX, y: secondNextKingPositionY });
        if (debugOn && king.color === debugCollor && kingSideCastle === debugKingSideCastle) console.log('canCastle: isPositionSafeA', isPositionSafeA)
        if (debugOn && king.color === debugCollor && kingSideCastle === debugKingSideCastle) console.log('canCastle: isPositionSafeB', isPositionSafeB)
        return isPositionSafeA && isPositionSafeB;
    }

    public move(from: Coords, to: Coords, promotedPieceType: FENChar | null): void {
        // Doing Move
        console.log('In move', from, to, promotedPieceType)

        // Check if the game is over
        if (this._isGameOver) {
            console.log('Game is over so bailing')
            throw new Error('Game is over');
        }

        // Check if coords are valid
        if (!this.areCoordsValid(from) || !this.areCoordsValid(to)) {
            console.log('Coords invalid so bailing')
            return;
        }


        // Check if the piece is valid
        const piece = this.chessBoard[from.x][from.y];
        if (!piece) {
            console.log('Piece not valid so bailing')
            return;
        }

        const piecesSafeSquares: Coords[] | undefined = this._safeSquares.get(from.x + "," + from.y);
        if (!piecesSafeSquares || !piecesSafeSquares.some(coords => coords.x === to.x && coords.y === to.y)) {
            console.log('Square is not safe to bailing')
            throw new Error('Square is not safe');
        }


        // If a Pawn or King or Rook update hasMoved
        if (piece instanceof Pawn || piece instanceof King || piece instanceof Rook) piece.hasMoved = true;

        const moveType = new Set<MoveType>();

        // Track if a piece is taken
        const isPieceTaken = this.chessBoard[to.x][to.y] !== null;
        if (isPieceTaken) moveType.add(MoveType.Capture);

        // Increment the fifty move rule counter
        if (piece instanceof Pawn || isPieceTaken) this.fiftyMoveRuleCounter = 0;
        else this.fiftyMoveRuleCounter += 0.5;

        // If castling, move the rook
        this.handlingSpecialMoves(piece, from, to, moveType);

        // Check for promotion
        if (promotedPieceType) {
            const promotedPiece = this.promotedPiece(promotedPieceType);
            this.chessBoard[to.x][to.y] = promotedPiece;
            this.chessBoard[from.x][from.y] = null;
            moveType.add(MoveType.Promotion);
        }
        else {
            // Update the board
            this.chessBoard[to.x][to.y] = piece;
            this.chessBoard[from.x][from.y] = null;
        }

        // Update the last move
        this._lastMove = { piece, from, to, moveType };

        // Swap the player color
        console.log('=============================================', this.playerColor)
        console.log('Swapping player color which is currently', this._playerColor)
        console.log('=============================================', this.playerColor)
        this._playerColor = this._playerColor === Color.White ? Color.Black : Color.White;

        // Get the safe squares and store them
        const safeSquares = this.findSafeSquares();
        
        // Check if the player is in check
        this.isInCheck(this._playerColor, true);

        // Track the move type
        if (this._checkState.isInChecked) {
             moveType.add(!(safeSquares.size === 0) ? MoveType.CheckMate : MoveType.Check);
        }
        else if (!moveType.size) {
            moveType.add(MoveType.BasicMove);
        }

        // Store the move
        this.storeMove(promotedPieceType);
        this.updateGameHistory();

        // We need to assign the safe squares after storing the move
        this._safeSquares = safeSquares;

        // Increment the full move counter
        if (this._playerColor === Color.White) this.fullNumberOfMoves++;

        // Update the board as FEN
        this._boardAsFEN = this.FENConverter.convertBoardToFEN(this.chessBoard, this._playerColor, this._lastMove, this.fiftyMoveRuleCounter, this.fullNumberOfMoves);
        this.updateThreeFoldRepetitionDictionary(this._boardAsFEN);

        // Check if the game is finished
        this._isGameOver = this.isGameFinished();
    }

    private handlingSpecialMoves(piece: Piece, from: Coords, to: Coords, moveTypes: Set<MoveType>): void {

        console.log('In handlingSpecialMoves', piece, from, to)
        // Handle Castling
        if (piece instanceof King && Math.abs(from.y - to.y) === 2) {

            console.log('Handling Castling')

            // Calculate rook from and to coords based on the direction of the move
            const rookFrom: Coords = { x: from.x, y: to.y === 6 ? 7 : 0 };
            const rookTo: Coords = { x: from.x, y: to.y === 6 ? 5 : 3 };

            // Check we have a rook
            const rook = this.chessBoard[rookFrom.x][rookFrom.y];
            if (!(rook instanceof Rook)) return;

            // Flag the rook as moved
            rook.hasMoved = true;

            // Move the rook
            this.chessBoard[rookTo.x][rookTo.y] = rook;
            this.chessBoard[rookFrom.x][rookFrom.y] = null;

            // Track the move type
            moveTypes.add(MoveType.Castling);
        }

        // Handle En Passant
        if (piece instanceof Pawn && this._lastMove) {
            const { piece: lastPiece, from: lastFrom, to: lastTo } = this._lastMove;
            if (
                lastPiece instanceof Pawn && // Last piece was a pawn
                Math.abs(lastFrom.x - lastTo.x) === 2 && // Last move was a two square move
                lastTo.y === to.y && // Last move was in the same column
                lastTo.x === from.x // Last move was next to the current pawn
            ) {
                this.chessBoard[lastTo.x][lastTo.y] = null;
                moveTypes.add(MoveType.Capture);
            }
        }
    }

    public promotedPiece(promotedPieceType: FENChar): Rook | Knight | Bishop | Queen {
        switch (promotedPieceType) {
            case FENChar.WhiteRook:
            case FENChar.BlackRook:
                return new Rook(this._playerColor);

            case FENChar.WhiteKnight:
            case FENChar.BlackKnight:
                return new Knight(this._playerColor);

            case FENChar.WhiteBishop:
            case FENChar.BlackBishop:
                return new Bishop(this._playerColor);

            case FENChar.WhiteQueen:
            case FENChar.BlackQueen:
                return new Queen(this._playerColor);

            default:
                throw new Error('Invalid piece type');
        }
    }

    private isGameFinished(): boolean {

        if (this.insufficientMaterial()) {
            this._gameOverMessage = "Draw due insufficient material";
            return true;
        }

        if (this._safeSquares.size === 0) {
            if (this._checkState.isInChecked) {
                const prevPlayer: string = this._playerColor === Color.White ? 'Black' : 'White';
                this._gameOverMessage = prevPlayer + ' wins by checkmate';
            }
            else {
                this._gameOverMessage = 'Draw by stalemate';
            }
            return true;
        }

        if (this.threeFoldRepetitionFlag) {
            this._gameOverMessage = "Draw due three fold repetition rule";
            return true;
        }

        if (this.fiftyMoveRuleCounter === 50) {
            this._gameOverMessage = "Draw due fifty move rule";
            return true;
        }

        return false;
    }


    private playerHasOnlyTwoKnightsAndKing(pieces: { piece: Piece, x: number, y: number }[]): boolean {
        return pieces.filter(piece => piece.piece instanceof Knight).length === 2;
    }

    private playerHasOnlyBishopsWithSameColorAndKing(pieces: { piece: Piece, x: number, y: number }[]): boolean {
        const bishops = pieces.filter(piece => piece.piece instanceof Bishop);
        const areAllBishopsOfSameColor = new Set(bishops.map(bishop => ChessBoard.isSquareDark(bishop.x, bishop.y))).size === 1;
        return bishops.length === pieces.length - 1 && areAllBishopsOfSameColor;
    }

    private insufficientMaterial(): boolean {
        const whitePieces: { piece: Piece, x: number, y: number }[] = [];
        const blackPieces: { piece: Piece, x: number, y: number }[] = [];

        for (let x = 0; x < ChessBoard.chessBoardSize; x++) {
            for (let y = 0; y < ChessBoard.chessBoardSize; y++) {
                const piece: Piece | null = this.chessBoard[x][y];
                if (!piece) continue;

                if (piece.color === Color.White) whitePieces.push({ piece, x, y });
                else blackPieces.push({ piece, x, y });
            }
        }

        // King vs King
        if (whitePieces.length === 1 && blackPieces.length === 1)
            return true;

        // King and Minor Piece vs King
        if (whitePieces.length === 1 && blackPieces.length === 2)
            return blackPieces.some(piece => piece.piece instanceof Knight || piece.piece instanceof Bishop);

        else if (whitePieces.length === 2 && blackPieces.length === 1)
            return whitePieces.some(piece => piece.piece instanceof Knight || piece.piece instanceof Bishop);

        // both sides have bishop of same color
        else if (whitePieces.length === 2 && blackPieces.length === 2) {
            const whiteBishop = whitePieces.find(piece => piece.piece instanceof Bishop);
            const blackBishop = blackPieces.find(piece => piece.piece instanceof Bishop);

            if (whiteBishop && blackBishop) {
                const areBishopsOfSameColor: boolean = ChessBoard.isSquareDark(whiteBishop.x, whiteBishop.y) && ChessBoard.isSquareDark(blackBishop.x, blackBishop.y) || !ChessBoard.isSquareDark(whiteBishop.x, whiteBishop.y) && !ChessBoard.isSquareDark(blackBishop.x, blackBishop.y);

                return areBishopsOfSameColor;
            }
        }

        if (whitePieces.length === 3 && blackPieces.length === 1 && this.playerHasOnlyTwoKnightsAndKing(whitePieces) ||
            whitePieces.length === 1 && blackPieces.length === 3 && this.playerHasOnlyTwoKnightsAndKing(blackPieces)
        ) return true;

        if (whitePieces.length >= 3 && blackPieces.length === 1 && this.playerHasOnlyBishopsWithSameColorAndKing(whitePieces) ||
            whitePieces.length === 1 && blackPieces.length >= 3 && this.playerHasOnlyBishopsWithSameColorAndKing(blackPieces)
        ) return true;

        return false;
    }

    private updateThreeFoldRepetitionDictionary(FEN: string): void {
        const threeFoldRepetitionFENKey: string = FEN.split(" ").slice(0, 4).join("");
        const threeFoldRepetionValue: number | undefined = this.threeFoldRepetitionDictionary.get(threeFoldRepetitionFENKey);

        if (threeFoldRepetionValue === undefined)
            this.threeFoldRepetitionDictionary.set(threeFoldRepetitionFENKey, 1);
        else {
            if (threeFoldRepetionValue === 2) {
                this.threeFoldRepetitionFlag = true;
                return;
            }
            this.threeFoldRepetitionDictionary.set(threeFoldRepetitionFENKey, 2);
        }
    }

    private startingPieceCoordsNotation(): string {
        const { piece: currPiece, from, to } = this._lastMove!;

        console.log('In startingPieceCoordsNotation', currPiece, from, to)

        // For King we only have one king. For Pawn this will be handled later
        if (currPiece instanceof Pawn || currPiece instanceof King) {
            console.log('Returning empty string 1')
            return "";
        }

        // For other pieces we need to get coordinates of all pieces of the same type that made the move
        const samePiecesCoords: Coords[] = [{ x: from.x, y: from.y }];

        // Iterate over the board to find the piece with the same type that made the move
        for (let x = 0; x < ChessBoard.chessBoardSize; x++) {
            for (let y = 0; y < ChessBoard.chessBoardSize; y++) {

                const piece: Piece | null = this.chessBoard[x][y];

                // Continue if the piece is null or the same piece
                if (!piece || (to.x === x && to.x === y)) {
                    console.log('Continuing')
                    continue;
                }

                if (piece.FENChar === currPiece.FENChar) {
                    const safeSquares: Coords[] = this._safeSquares.get(x + "," + y) || [];
                    const pieceHasSameTargetSquare: boolean = safeSquares.some(coords => coords.x === to.x && coords.y === to.y);
                    if (pieceHasSameTargetSquare) {
                        samePiecesCoords.push({ x, y });
                    }
                }
            }
        }

        if (samePiecesCoords.length === 1) {
            console.log('Returning empty string 2')
            return "";
        }

        const piecesFile = new Set(samePiecesCoords.map(coords => coords.y));
        const piecesRank = new Set(samePiecesCoords.map(coords => coords.x));

        // means that all of the pieces are on different files (a, b, c, ...)
        if (piecesFile.size === samePiecesCoords.length) {
            const result = columns[from.y];
            console.log('Returning result', result)
            return result
        }

        // means that all of the pieces are on different rank (1, 2, 3, ...)
        if (piecesRank.size === samePiecesCoords.length) {
            const result = String(from.x + 1);
            console.log('Returning result', result)
            return result;
        }

        // in case that there are pieces that shares both rank and a file with multiple or one piece
        const result =columns[from.y] + String(from.x + 1);
        console.log('Returning result', result)
        return result;
    }

    private updateGameHistory(): void {
        this._gameHistory.push({
            board: [...this.chessBoardView.map(row => [...row])],
            checkState: { ...this._checkState },
            lastMove: this._lastMove ? { ...this._lastMove } : undefined
        });
    }

    private storeMove(promotedPiece: FENChar | null, longNotation=false): void {

        // Destructure the last move
        const { piece, from, to, moveType } = this._lastMove!;

        // Set name of the piece
        let pieceName: string = !(piece instanceof Pawn) ? piece.FENChar.toUpperCase() : "";

        let move: string;

        // Check if the move is castling
        if (moveType.has(MoveType.Castling)) {

            // Check if the move is kingside or queenside
            move = to.y - from.y === 2 ? "O-O" : "O-O-O";
            console.log('Assign move 0', move)
        }
        else {
            move = longNotation ? 
                pieceName + columns[from.y] + String(from.x + 1) :
                pieceName + this.startingPieceCoordsNotation();
            console.log('Assign move 1', move)

            // If capturing, add an x
             if (moveType.has(MoveType.Capture)) {
                // Long notation
                if(longNotation) {
                    move += "x";
                    console.log('Assign move 2', move)
                }
                // Short notation
                else {
                    // Check if the piece is a pawn add y coordinate of the piece and x
                    if(piece instanceof Pawn) {
                        move += columns[from.y] + "x";
                        console.log('Assign move 2.1', move)
                    }
                    else {
                        move += "x";
                        console.log('Assign move 2.2', move)
                    }
                }
            } 

            // Add the target square to move
            move += columns[to.y] + String(to.x + 1);
            console.log('Assign move 3', move)

            // Check if the move is a promotion and add the promoted piece
            if (promotedPiece) {
                move += "=" + promotedPiece.toUpperCase();
                console.log('Assign move 4', move)
            }
        }

        // Check if the move is a check append a +, 
        if (moveType.has(MoveType.Check)) {
            move += "+";
            console.log('Assign move 5', move)
        }
        // if checkmate append a #
        else if (moveType.has(MoveType.CheckMate)) {
            move += "#";
            console.log('Assign move 6',move)
        }

        console.log('Move is', move)
        console.log('fullNumberOfMoves',this.fullNumberOfMoves - 1)

        if (!this._moveList[this.fullNumberOfMoves - 1])
            this._moveList[this.fullNumberOfMoves - 1] = [move];
        else
            this._moveList[this.fullNumberOfMoves - 1].push(move);
    }

}