import { Component, OnDestroy, OnInit } from '@angular/core';
import { ChessBoard } from '../../chess-logic/chess-board';
import { CheckState, Color, Coords, FENChar, GameHistory, LastMove, MoveList, MoveType, pieceImagePaths, SafeSquares, SelectedSquare } from '../../chess-logic/models';
import { CommonModule, NgClass, NgFor, NgIf } from '@angular/common';
import { ChessBoardService } from './chess-board.service';
import { MoveListComponent } from '../move-list/move-list.component';
import { filter, from, fromEvent, Subscription, tap } from 'rxjs';

@Component({
  selector: 'app-chess-board',
  standalone: true,
  imports: [MoveListComponent, CommonModule],
  templateUrl: './chess-board.component.html',
  styleUrl: './chess-board.component.css'
})
export class ChessBoardComponent implements OnInit, OnDestroy {
  public pieceImagePaths = pieceImagePaths

  // The chess board is a private member of the component
  protected chessBoard =  new ChessBoard();

  // The chess board view is a public member of the component
  public chessBoardView : (FENChar|null)[][] = this.chessBoard.chessBoardView;

  // The player color is a public member of the component
  public get playerColor(): Color { return this.chessBoard.playerColor; };

  // The safe squares is a public member
  public get safeSquares(): SafeSquares { return this.chessBoard.safeSquares; }

  // The game over message is a public member
  public get gameOverMessage(): string | undefined { return this.chessBoard.gameOverMessage; }

  private selectedSquare: SelectedSquare = {piece: null};
  private pieceSafeSquares: Coords[] = [];


  private lastMove:LastMove| undefined = this.chessBoard.lastMove;
  private checkState: CheckState = this.chessBoard.checkState;

  public get moveList(): MoveList { return this.chessBoard.moveList;  }
  public get gameHistory(): GameHistory { return this.chessBoard.gameHistory; };
  public gameHistoryPointer: number = 0;
  
  // Promotion properties
  public isPromotionModalVisible: boolean = false;
  private promotionCoords: Coords | null = null;
  private promotedPiece: FENChar | null = null;

  public promotionalPieces(): FENChar[] {
    return (this.playerColor === Color.White ?
      [FENChar.WhiteQueen, FENChar.WhiteRook, FENChar.WhiteBishop, FENChar.WhiteKnight] :
      [FENChar.BlackQueen, FENChar.BlackRook, FENChar.BlackBishop, FENChar.BlackKnight]);
  }

  public flipMode = false;

  private subscriptions$ = new Subscription();  
  
  constructor(protected chessBoardService: ChessBoardService) { }

  public flipBoard(): void {
    this.flipMode = !this.flipMode;
    this.chessBoardService.chessBoardState$.next(this.chessBoard.boardAsFEN);
    console.log('flipMode', this.flipMode);
  }

  public isSquareDark(row: number, col: number): boolean {
    return ChessBoard.isSquareDark(row, col);
  }

  public isSquareSelected(row: number, col: number): boolean {
    if(!this.selectedSquare.piece) return false
    return this.selectedSquare.coords.x === row && this.selectedSquare.coords.y === col;
  }

  public isSquareSafeForSelectedPiece(row: number, col: number): boolean {
    return this.pieceSafeSquares.some(coords => coords.x === row && coords.y === col);
  }

  public isSquareLastMove(row: number, col: number): boolean {
    if (!this.lastMove) return false;
    return this.lastMove.to.x === row && this.lastMove.to.y === col || 
           this.lastMove.from.x === row && this.lastMove.from.y === col;
  }

  public isSquareChecked(row: number, col: number): boolean {
    const result = this.checkState.isInChecked && this.checkState.coords.x === row && this.checkState.coords.y === col;
    if(this.checkState.isInChecked && result) {
      console.log('---------------------------------')
      console.log('incoming row', row);
      console.log('incoming col', col);
      console.log('checkState', this.checkState);
      console.log('checkState.isInChecked', this.checkState.isInChecked);
      console.log('checkState coords', this.checkState.coords);
      console.log('row', this.checkState.coords.x);
      console.log('col', this.checkState.coords.y);
      console.log('result', result);
    }

    // console.log('result', result);
    return result
  }
    
  public isSquarePromotionSquare(row: number, col: number): boolean {
    if (!this.promotionCoords) return false;
    return this.promotionCoords.x === row && this.promotionCoords.y === col;
  }

  public unmarkingPreviouslySelectedAndfSafeSquares(): void {
    this.selectedSquare = {piece: null};
    this.pieceSafeSquares = [];

    if(this.isPromotionModalVisible) {
      this.isPromotionModalVisible = false;
      this.promotedPiece = null;
      this.promotionCoords = null;
    }

  }

  public selectingPiece(row: number, col: number): void {
    if(this.gameOverMessage) return;

    console.log('Selecting', row, col);
    const piece :FENChar | null = this.chessBoardView[row][col];
    if (!piece) return
    if (this.isWrongPieceSelected(piece)) return;

    const isSameSquareClicked: boolean = !!this.selectedSquare.piece && this.selectedSquare.coords.x === row && this.selectedSquare.coords.y === col;
    if (isSameSquareClicked) {
      this.unmarkingPreviouslySelectedAndfSafeSquares();
      return;
    }

    this.selectedSquare = {piece: piece, coords: {x: row, y: col}};
    this.pieceSafeSquares = this.safeSquares.get(row + "," + col) || [];
  }

  private placingPiece(row: number, col: number): void {
    if (!this.selectedSquare.piece) return;
    if (!this.isSquareSafeForSelectedPiece(row, col)) return;

    // Pawn promotion
    const shouldOpenDialog = 
        (this.selectedSquare.piece === FENChar.WhitePawn && row === 7 ||
        this.selectedSquare.piece === FENChar.BlackPawn && row === 0) 

    if(shouldOpenDialog) {

      // Clear the selected square and safe squares if we are promoting a pawn
      this.pieceSafeSquares = [];

      // Open the promotion modal
      this.isPromotionModalVisible = true;

      // Set the promotion coords
      this.promotionCoords = {x: row, y: col};

      return;
    } 

    this.updateBoard(this.selectedSquare.coords, {x: row, y: col}, this.promotedPiece);
  }

  protected updateBoard(from: Coords, to: Coords, promotedPiece: FENChar| null): void {
    this.chessBoard.move(from, to, promotedPiece);
    this.chessBoardView = this.chessBoard.chessBoardView;
    this.checkState = this.chessBoard.checkState;
    this.lastMove = this.chessBoard.lastMove;
    this.markLastMoveAndCheckState(this.lastMove, this.chessBoard.checkState);
    this.unmarkingPreviouslySelectedAndfSafeSquares();
    console.log('=============================================', this.playerColor)
    console.log('Updating next FEN position', this.chessBoard.boardAsFEN);
    this.chessBoardService.chessBoardState$.next(this.chessBoard.boardAsFEN);
    this.gameHistoryPointer++;
  }

  public promotePiece(piece: FENChar): void {
    if (!this.promotionCoords) return; // Check we have a promotion coords
    if (!this.selectedSquare.piece) return; // Check we have a selected piece

    this.promotedPiece = piece;
    this.updateBoard(this.selectedSquare.coords, this.promotionCoords, this.promotedPiece);
  }

  public closePromotionDialog():void {
    this.unmarkingPreviouslySelectedAndfSafeSquares();
  }
 
  private markLastMoveAndCheckState(lastMove: LastMove | undefined, checkState: CheckState): void {
    this.lastMove = lastMove;
    this.checkState = checkState;

    if (this.lastMove) {
      this.moveSound(this.lastMove.moveType);
    }
    else {
      this.moveSound(new Set<MoveType>([MoveType.BasicMove ]));
    } 
  }

  public move(row: number, col: number): void {
    this.selectingPiece(row, col);
    this.placingPiece(row, col);
  }

  private isWrongPieceSelected(piece: FENChar): boolean {
    const isWhitePieceSelected: boolean = piece === piece.toUpperCase();
    return isWhitePieceSelected && this.playerColor === Color.Black ||
      !isWhitePieceSelected && this.playerColor === Color.White;
  }

  public showPreviousPosition(moveIndex: number): void {
    const { board, checkState, lastMove } = this.gameHistory[moveIndex];
    this.chessBoardView = board;
    this.markLastMoveAndCheckState(lastMove, checkState);
    this.gameHistoryPointer = moveIndex;
  }

  private moveSound(moveType: Set<MoveType>): void {
    const moveSound = new Audio("/assets/sounds/move.mp3");
    
    if (moveType.has(MoveType.Promotion)) moveSound.src = "assets/sounds/promotion.mp3";
    else if (moveType.has(MoveType.Capture)) moveSound.src = "assets/sounds/capture.mp3";
    else if (moveType.has(MoveType.Castling)) moveSound.src = "assets/sounds/castling.mp3";
    
    if (moveType.has(MoveType.CheckMate)) moveSound.src = "assets/sounds/checkmate.mp3";
    else if (moveType.has(MoveType.Check)) moveSound.src = "assets/sounds/check.mp3";

    console.log('Playing sound', moveSound.src);
    moveSound.play();
  }

  ngOnInit(): void {
    const keyboardSubscription$ = fromEvent<KeyboardEvent>(document, 'keyup')
    .pipe(
      filter((event: KeyboardEvent) => event.key === 'ArrowLeft' || event.key === 'ArrowRight'),
      tap((event: KeyboardEvent) => {
        switch(event.key) {
          case 'ArrowRight':
            if (this.gameHistoryPointer === this.gameHistory.length - 1) return;
            this.gameHistoryPointer++ ;
            break;
          case 'ArrowLeft':
            if (this.gameHistoryPointer === 0) return;
            this.gameHistoryPointer--;
            break;
          default:
            break;
        }

        this.showPreviousPosition(this.gameHistoryPointer); 
      })
    ).subscribe();

    this.subscriptions$.add(keyboardSubscription$);
  }

  ngOnDestroy(): void {
    this.subscriptions$.unsubscribe();
  }
}
